@@ -20,7 +20,7 @@ TODO:
    (imported modules ? use dis.dis ?)
 
 
-:version:   $Revision: 1.36 $  
+:version:   $Revision: 1.54 $  
 :author:    Sylvain Thenault
 :copyright: 2003-2005 LOGILAB S.A. (Paris, FRANCE)
 :contact:   http://www.logilab.fr/ -- mailto:python-projects@logilab.org
@@ -28,14 +28,13 @@ TODO:
 :contact:   mailto:thenault@gmail.com
 """
 
-__revision__ = "$Id: builder.py,v 1.36 2005/11/02 17:19:25 syt Exp $"
+__revision__ = "$Id: builder.py,v 1.54 2006-03-14 15:21:33 syt Exp $"
 __doctype__ = "restructuredtext en"
 
 import sys
 from os.path import splitext, basename, dirname, exists, abspath
 from parser import ParserError
 from compiler import parse
-from compiler.ast import EmptyNode
 from inspect import isfunction, ismethod, ismethoddescriptor, isclass, \
      isbuiltin
 try: # python 2.2 inspect module doesn't have the isdatadescriptor function
@@ -51,9 +50,7 @@ from logilab.common.modutils import modpath_from_file
 from logilab.astng import nodes
 from logilab.astng.utils import ASTWalker
 from logilab.astng._exceptions import ASTNGBuildingException
-from logilab.astng.raw_building import register_arguments,  build_module, \
-     object_build_class, object_build_function, \
-     object_build_datadescriptor, object_build_methoddescriptor
+from logilab.astng.raw_building import *
 
 # ast NG builder ##############################################################
 
@@ -61,12 +58,14 @@ class ASTNGBuilder:
     """provide astng building methods
     """
     
-    def __init__(self, manager):
-        self._manager = manager
+    def __init__(self): # XXX _ was the manager, keep for bw compat
         self._module = None
+        self._file = None
         self._done = None
         self._stack, self._par_stack = None, None
+        self._metaclass = None        
         self._walker = ASTWalker(self)
+        self._dyn_modname_map = {'gtk': 'gtk._gtk'}
         
     def module_build(self, module, modname=None):
         """build an astng from a living module instance
@@ -81,18 +80,19 @@ class ASTNGBuilder:
         if node is None:
             # this is a built-in module
             # get a partial representation by introspection
-            node = self.inspect_build(module)
-            node.file = node.path = path and abspath(path) or path
-            node.package = hasattr(module, '__path__')
+            node = self.inspect_build(module, modname=modname, path=path)
         return node
 
-    def inspect_build(self, module, modname=None):
+    def inspect_build(self, module, modname=None, path=None):
         """build astng from a living module (i.e. using inspect)
         this is used when there is no python source code available (either
         because it's a built-in module or because the .py is not available)
         """
         self._module = module
         node = build_module(modname or module.__name__, module.__doc__)
+        node.file = node.path = path and abspath(path) or path
+        node.package = hasattr(module, '__path__')
+        attach___dict__(node)
         self._done = {}
         self.object_build(node, module)
         return node
@@ -107,17 +107,20 @@ class ASTNGBuilder:
         except IOError, ex:
             msg = 'Unable to load file %r (%s)' % (path, ex)
             raise ASTNGBuildingException(msg)
+        self._file = path
         # get module name if necessary, *before modifying sys.path*
         if modname is None:
             try:
                 modname = '.'.join(modpath_from_file(path))
-            except:
+            except ImportError:
                 modname = splitext(basename(path))[0]
         # build astng representation
         try:
             sys.path.insert(0, dirname(path))
             node = self.string_build(data, modname, path)
+            node.file = abspath(path)
         finally:
+            self._file = None
             sys.path.pop(0)
         return node
     
@@ -157,6 +160,8 @@ class ASTNGBuilder:
         """
         self._stack = [self._module]
         self._par_stack = [node]
+        self._metaclass = ['']
+        self._global_names = []
         node.parent = None
         node.globals = node.locals = {}
         for name, value in ( ('__name__', node.name),
@@ -165,6 +170,7 @@ class ASTNGBuilder:
             const = nodes.Const(value)
             const.parent = node
             node.locals[name] = [const]
+        attach___dict__(node)
         if node.package:
             # FIXME: List(Const())
             const = nodes.Const(dirname(node.path))
@@ -172,7 +178,6 @@ class ASTNGBuilder:
             node.locals['__path__'] = [const]
             
 
-
     def leave_module(self, _):
         """leave a stmt.Module node -> pop the last item on the stack and check
         the stack is empty
@@ -191,19 +196,32 @@ class ASTNGBuilder:
         node.instance_attrs = {}
         node.basenames = [b_node.as_string() for b_node in node.bases]
         self._push(node)
-        node.locals['__doc__'] = [node.doc]
+        for name, value in ( ('__name__', node.name),
+                             ('__module__', node.root().name),
+                             ('__doc__', node.doc) ):
+            const = nodes.Const(value)
+            const.parent = node
+            node.locals[name] = [const]
+        attach___dict__(node)
+        self._metaclass.append(self._metaclass[-1])
         
     def leave_class(self, node):
         """leave a stmt.Class node -> pop the last item on the stack
         """
         self.leave_default(node)
         self._stack.pop()
+        metaclass = self._metaclass.pop()
+        if not node.bases:
+            # no base classes, detect new / style old style according to
+            # current scope
+            node._newstyle = metaclass == 'type'
         
     def visit_function(self, node):
         """visit a stmt.Function node -> init node and push the corresponding
         object or None on the top of the stack
         """
         self.visit_default(node)
+        self._global_names.append({})
         node.argnames = list(node.argnames)
         if isinstance(node.parent.frame(), nodes.Class):
             node.type = 'method'
@@ -217,6 +235,7 @@ class ASTNGBuilder:
         """
         self.leave_default(node)
         self._stack.pop()
+        self._global_names.pop()
         
     def visit_lambda(self, node):
         """visit a stmt.Lambda node -> init node locals
@@ -230,8 +249,15 @@ class ASTNGBuilder:
         """visit a stmt.Global node -> add declared names to locals
         """
         self.visit_default(node)
+        if not self._global_names: # global at the module level, no effect
+            return
         for name in node.names:
-            node.parent.set_local(name, node)
+            self._global_names[-1].setdefault(name, []).append(node)
+#             node.parent.set_local(name, node)
+#         module = node.root()
+#         if module is not node.frame():
+#             for name in node.names:
+#                 module.set_local(name, node)
             
     def visit_import(self, node):
         """visit a stmt.Import node -> add imported names to locals
@@ -245,14 +271,12 @@ class ASTNGBuilder:
         """visit a stmt.From node -> add imported names to locals
         """
         self.visit_default(node)
-        # add names imported by wildcard import
+        # add names imported by the import to locals
         for (name, asname) in node.names:
             if name == '*':
                 try:
-                    imported = self._manager.astng_from_module_name(node.modname)
-                    for name in imported.wildcard_import_names():
-                        node.parent.set_local(name, node)
-                except:
+                    imported = node.root().import_module(node.modname)
+                except ASTNGBuildingException:
                     #import traceback
                     #traceback.print_exc()
                     continue
@@ -260,6 +284,8 @@ class ASTNGBuilder:
                     #print >> sys.stderr, \
                     #      'Unable to get imported names for %r line %s"' % (
                     #    node.modname, node.lineno)
+                for name in imported.wildcard_import_names():
+                    node.parent.set_local(name, node)
             else:
                 node.parent.set_local(asname or name, node)
 
@@ -276,6 +302,7 @@ class ASTNGBuilder:
         
     def visit_assign(self, node):
         """visit a stmt.Assign node -> check for classmethod and staticmethod
+        + __metaclass__
         """
         self.visit_default(node)
         klass = node.parent.frame()
@@ -296,20 +323,35 @@ class ASTNGBuilder:
                         except KeyError:
                             #print >> sys.stderr, 'FIXME 2', ass_node.name
                             continue
+        elif (isinstance(node.nodes[0], nodes.AssName)
+              and node.nodes[0].name == '__metaclass__'): # XXX check more...
+            self._metaclass[-1] = 'type'
 
     def visit_assname(self, node):
         """visit a stmt.AssName node -> add name to locals
         """
         self.visit_default(node)
-        if node.flags == 'OP_ASSIGN':
-            node.parent.set_local(node.name, node)
-            
+        self._add_local(node, node.name)
+
+    def visit_augassign(self, node):
+        """visit a stmt.AssName node -> add name to locals
+        """
+        self.visit_default(node)
+        if not isinstance(node.node, nodes.Name):
+            return  # XXX
+        self._add_local(node, node.node.name)
+
+    def _add_local(self, node, name):
+        if self._global_names and name in self._global_names[-1]:
+            node.root().set_local(name, node)
+        else:
+            node.parent.set_local(name, node)
+        
     def visit_assattr(self, node):
         """visit a stmt.AssAttr node -> add name to locals, handle members
         definition
         """
         self.visit_default(node)
-        #node.parent.set_local(node.name, node)
         frame = node.frame()
         if isinstance(frame, nodes.Function) and frame.type != 'function':
             klass = frame.parent.frame()
@@ -320,15 +362,26 @@ class ASTNGBuilder:
                 # first argument is missing !
                 return
             if isinstance(node.expr, nodes.Name) and node.expr.name == _self:
-                iattrs = klass.instance_attrs
+                # unittest_scoped_nodes.ClassNodeTC.test_classmethod_attributes
+                #
+                # if frame.type == 'classmethod': XXX at this point we may have
+                # not encountered the classmethod decorator, so we havn't yet
+                # the correct type
+                # hack according to the argument name
+                if _self == 'self': 
+                    iattrs = klass.instance_attrs
+                else:
+                    iattrs = klass.locals
                 # assign if not yet existant in others
                 if not iattrs.has_key(node.attrname):
-                    iattrs[node.attrname] = node
+                    iattrs[node.attrname] = [node]
                 # but always assign in __init__, except if previous assigment
                 # already come from __init__
                 elif frame.name == '__init__' and not \
-                         iattrs[node.attrname].frame().name == '__init__':
-                    klass.instance_attrs[node.attrname] = node
+                         iattrs[node.attrname][0].frame().name == '__init__':
+                    iattrs[node.attrname].insert(0, node)
+                else:
+                    iattrs[node.attrname].append(node)
                     
     def visit_default(self, node):
         """default visit method, handle the parent attribute
@@ -353,7 +406,7 @@ class ASTNGBuilder:
     # astng from living objects ###############################################
     #
     # this is actually a really minimal representation, including only Module,
-    # Function and Class nodes
+    # Function and Class nodes and some others as guessed
     
     def object_build(self, node, obj):
         """recursive method which create a partial ast from real objects
@@ -379,23 +432,16 @@ class ASTNGBuilder:
                     attach_dummy_node(node, name)
                     continue
                 object_build_function(node, member)
-            elif ismethoddescriptor(member):
-                object_build_methoddescriptor(node, member)
-            elif isdatadescriptor(member):
-                object_build_datadescriptor(node, member)
             elif isbuiltin(member):
                 # verify this is not an imported member
-                if getattr(member, '__module__', None) != modname:
-                    attach_dummy_node(node, name)
+                if self._member_module(member) != modname:
+                    imported_member(node, member, name)
                     continue
                 object_build_methoddescriptor(node, member)                
             elif isclass(member):
                 # verify this is not an imported class
-                #
-                # /!\ some classes like ExtensionClass doesn't have a 
-                # __module__ attribute !
-                if getattr(member, '__module__', None) != modname:
-                    attach_dummy_node(node, name)
+                if self._member_module(member) != modname:
+                    imported_member(node, member, name)
                     continue
                 if member in self._done:
                     class_node = self._done[member]
@@ -404,16 +450,36 @@ class ASTNGBuilder:
                     class_node = object_build_class(node, member)
                 # recursion
                 self.object_build(class_node, member)
+            elif ismethoddescriptor(member):
+                assert isinstance(member, object)
+                object_build_methoddescriptor(node, member)
+            elif isdatadescriptor(member):
+                assert isinstance(member, object)
+                object_build_datadescriptor(node, member, name)
+            elif isinstance(member, (int, long, float, str, unicode)) or member is None:
+                attach_const_node(node, name, member)
             else:
                 # create an empty node so that the name is actually defined
                 attach_dummy_node(node, name)
 
-def attach_dummy_node(node, name):
-    dummy = EmptyNode()
-    dummy.name = name
-    dummy.parent = node
-    dummy.lineno = 1
-    node.add_local_node(dummy)
+    def _member_module(self, member):
+        modname = getattr(member, '__module__', None)
+        return self._dyn_modname_map.get(modname, modname)
+        
+def imported_member(node, member, name):
+    """consider a class/builtin member where __module__ != current module name
+
+    check if it's sound valid and then add an import node, else use a dummy node
+    """
+    # /!\ some classes like ExtensionClass doesn't have a 
+    # __module__ attribute !
+    member_module = getattr(member, '__module__', '__builtin__')
+    try:
+        getattr(sys.modules[member_module], name)
+    except (KeyError, AttributeError):
+        attach_dummy_node(node, name)
+    else:
+        attach_import_node(node, member_module, name)
     
 # optimize the tokenize module
 #from logilab.common.bind import optimize_module
