@@ -20,9 +20,6 @@ on all nodes :
  .statement(), returning the first parent node marked as statement node
  .frame(), returning the first node defining a new local scope (i.e.
   Module, Function or Class)
- .resolve(name), try to resolve the value (astng node) associated to an
-  identifier <name>. Raises ResolveError if it has been unable to resolve the
-  name.
  .set_local(name, node), define an identifier <name> on the first parent frame,
   with the node defining it. This is used by the astng builder and should not
   be used from out there.
@@ -34,23 +31,21 @@ on From and Import :
 
  [1] http://docs.python.org/lib/module-compiler.ast.html
 
-:version:   $Revision: 1.81 $  
+:version:   $Revision: 1.103 $  
 :author:    Sylvain Thenault
-:copyright: 2003-2005 LOGILAB S.A. (Paris, FRANCE)
+:copyright: 2003-2006 LOGILAB S.A. (Paris, FRANCE)
 :contact:   http://www.logilab.fr/ -- mailto:python-projects@logilab.org
-:copyright: 2003-2005 Sylvain Thenault
+:copyright: 2003-2006 Sylvain Thenault
 :contact:   mailto:thenault@gmail.com
 """
 
 from __future__ import generators
 
-__revision__ = "$Id: nodes.py,v 1.81 2005/11/07 14:02:56 syt Exp $"
+__revision__ = "$Id: nodes.py,v 1.103 2006-04-19 14:31:38 syt Exp $"
 __doctype__ = "restructuredtext en"
 
-import __builtin__
-from inspect import isclass
-from compiler.ast import Assign, Add, And, AssAttr, AssList, AssName, AssTuple, \
-     Assert, Assign, AugAssign, \
+from compiler.ast import Assign, Add, And, AssAttr, AssList, AssName, \
+     AssTuple, Assert, Assign, AugAssign, \
      Backquote, Bitand, Bitor, Bitxor, Break, CallFunc, Class, \
      Compare, Const, Continue, Dict, Discard, Div, FloorDiv, \
      Ellipsis, EmptyNode, Exec, \
@@ -61,20 +56,26 @@ from compiler.ast import Assign, Add, And, AssAttr, AssList, AssName, AssTuple,
      Sliceobj, Stmt, Sub, Subscript, TryExcept, TryFinally, Tuple, UnaryAdd, \
      UnarySub, While, Yield
 try:
-    from compiler.ast import GenExpr, GenExprFor
+    from compiler.ast import GenExpr, GenExprFor, GenExprIf, GenExprInner
 except:
-    pass
+    class GenExpr:
+        """dummy GenExpr node, shouldn't be used since py < 2.4"""
+    class GenExprFor: 
+        """dummy GenExprFor node, shouldn't be used since py < 2.4"""
+    class GenExprIf: 
+        """dummy GenExprIf node, shouldn't be used since py < 2.4"""
+    class GenExprInner: 
+        """dummy GenExprInner node, shouldn't be used since py < 2.4"""
 
-from logilab.astng import ASTNGManager
-from logilab.astng._exceptions import ResolveError, NotFoundError, \
-     ASTNGBuildingException
+from logilab.astng._exceptions import NotFoundError, InferenceError
 from logilab.astng.utils import extend_class
-from logilab.astng.scoped_nodes import LocalsDictMixIn
 
 import re
 ID_RGX = re.compile('^[a-zA-Z_][a-zA-Z_0-9]*$')
 del re
 
+INFER_NEED_NAME_STMTS = (From, Import, Global, TryExcept)
+
 # Node  ######################################################################
 
 class NodeNG:
@@ -88,7 +89,19 @@ class NodeNG:
 
     # parent node in the tree
     parent = None
+
+    def __str__(self):
+        return '%s(%s)' % (self.__class__.__name__, getattr(self, 'name', ''))
     
+    def parent_of(self, node):
+        """return true if i'm a parent of the given node"""
+        parent = node.parent
+        while parent is not None:
+            if self is parent:
+                return True
+            parent = parent.parent
+        return False
+
     def is_statement(self):
         """return true if the node should be considered as statement node
         """
@@ -145,7 +158,8 @@ class NodeNG:
         mylineno = self.source_line()
         nearest = None, 0
         for node in nodes:
-            assert node.root() is myroot, 'not from the same module %s' % (self, node)
+            assert node.root() is myroot, \
+                   'not from the same module %s' % (self, node)
             lineno = node.source_line()
             if node.source_line() > mylineno:
                 break
@@ -157,7 +171,7 @@ class NodeNG:
     def source_line(self):
         """return the line number where the given node appears
 
-        we need this method since not all node as the lineno attribute
+        we need this method since not all nodes as the lineno attribute
         correctly set...
         """
         line = self.lineno
@@ -174,13 +188,32 @@ class NodeNG:
                     _node = _node.parent
             self.lineno = line
         return line
+    
+    def last_source_line(self):
+        """return the last block line number for this node (i.e. including
+        children)
+        """
+        try:
+            return self.__dict__['_cached_last_source_line']
+        except KeyError:
+            line = self.source_line()
+            for node in self.getChildNodes():
+                line = max(line, node.last_source_line())
+            self._cached_last_source_line = line
+            return line
+
+    def block_range(self, lineno):
+        """handle block line numbers range for non block opening statements
+        """
+        return lineno, self.last_source_line()
+
 
     def set_local(self, name, stmt):
         """delegate to a scoped parent handling a locals dictionary
         """
         self.parent.set_local(name, stmt)
 
-    def nodes_of_class(self, klass):
+    def nodes_of_class(self, klass, skip_klass=None):
         """return an iterator on nodes which are instance of the given class(es)
 
         klass may be a class object or a tuple of class objects
@@ -188,157 +221,74 @@ class NodeNG:
         if isinstance(self, klass):
             yield self
         for child_node in self.getChildNodes():
+            if skip_klass is not None and isinstance(child_node, skip_klass):
+                continue
             for matching in child_node.nodes_of_class(klass):
                 yield matching
 
-    def names(self):
-        """return the list of accessed names from node"""
-        return [name.name for name in self.nodes_of_class(Name)]
-
-    def resolve_dotted(self, name):
-        """resolve a dotted names"""
-        parts = name.split('.')
-        baseobj = self.resolve(parts[0])
-        for part in parts[1:]:
-            # can't use isinstance since we don't use actual inheritance
-            #if not isinstance(baseobj, LocalsDictMixIn):
-            if not hasattr(baseobj, 'locals'):
-                raise ResolveError(name)
-            baseobj = baseobj.resolve(part)
-        return baseobj
-
-    def resolve_all(self, names):
-        """return an iterator the resolved names"""
-        for name in names:
-            try:
-                yield self.resolve_dotted(name)
-            except (ResolveError, NotFoundError, AssertionError):
-                continue
-
-    def resolve(self, name):
-        """try to resolve the value (astng node) associated to an
-        identifier <name>.
-
-        Raises ResolveError if it has been unable to resolve the name.
-
-        FIXME: refactor !!
-        """
-        #print 'resolve', self.__class__, getattr(self, 'name', '?'), name
-        assert ID_RGX.match(name), '%r is not a valid identifier' % name
-        frame = self.frame()
-        try:
-            stmts = frame.locals[name]
-        except KeyError:
-            try:
-                stmts = frame.root().locals[name]
-                # update the frame so that the correct one is given to self_resovle
-                frame = frame.root()
-            except KeyError:
-                if isinstance(frame, Module) and frame.package:
-                    modname = '%s.%s' % (frame.name, name)
-                    try:
-                        return ASTNGManager().astng_from_module_name(modname)
-                    except KeyboardInterrupt:
-                        raise
-                    except:
-                        pass
-                try:
-                    obj = getattr(__builtin__, name)
-                except AttributeError:
-                    raise ResolveError(name)
-                if name in ('None', 'True', 'False'):
-                    val_node = Const(eval(name))
-                    val_node.parent = ASTNGManager().astng_from_module(__builtin__)
-                    return val_node
-                module = getattr(obj, '__module__', '__builtin__')
-                # the first test avoid infinite loops
-                if frame.name.split('.')[-1] == module or not isclass(obj):
-                    raise ResolveError(name)
-                # FIXME: may not be a class
-                return ASTNGManager().astng_from_class(obj, module)
-        #print self.__class__, getattr(self, 'name', 'noname'), name,
-        for stmt in stmts:
-            #print stmt.__class__
-            try:
-                return stmt.self_resolve(name, frame)
-            except ResolveError:
-                continue
-        raise ResolveError(name)
-
-    def self_resolve(self, *_):
-        """self resolve return the node itself by default"""
-        return self
+    def _infer_name(self, frame, name):
+        if isinstance(self, INFER_NEED_NAME_STMTS) or (
+                 isinstance(self, (Function, Lambda)) and self is frame):
+            return name
+        return None
+    
 
 extend_class(Node, NodeNG)
 
+# block range overrides #######################################################
 
-# others... ###################################################################
-
-def assname_self_resolve(self, name, _):
-    """self resolve on AssName try to return the assigned value"""
-    # FIXME
-    assigned = self.rhs() or self
-    # FIXME: check name != assigned.name to avoid infinite loop
-    # if this is the case, we probably want to look for that name
-    # in globals/builtins, but this is not easlily feasible right now...
-    if isinstance(assigned, Name) and name != assigned.name:
-        return self.frame().resolve(assigned.name)
-    return assigned
-AssName.self_resolve = assname_self_resolve
-
-def _imported_module_astng(node, modname):
-    """return the ast for a module whose name is <modname> imported by <node>
+def function_block_range(node, lineno):
+    """handle block line numbers range for function statements
     """
-    # handle special case where we are on a package node importing a module
-    # using the same name as the package, which may end in an infinite loop
-    # on relative imports
-    if modname == getattr(node.root(), 'name', None):
-        # FIXME: I don't know what to do here...
-        raise ResolveError(name)
-    context_file = node.root().file
-    try:
-        return ASTNGManager().astng_from_module_name(modname, context_file)
-    except (ASTNGBuildingException, SyntaxError):
-        # FIXME
-        raise ResolveError(modname)
-    
-def import_self_resolve(self, name, _=None, asname=True):
-    """self resolve on From / Import nodes return the imported module/object"""
-    if asname:
-        return _imported_module_astng(self, self.real_name(name))
-    return _imported_module_astng(self, name)
-    
+    return node.source_line(), node.last_source_line()
 
-Import.self_resolve = import_self_resolve
+Function.block_range = function_block_range
 
-def from_self_resolve(self, name, _=None):
-    """self resolve on From / Import nodes return the imported module/object"""
-    module = _imported_module_astng(self, self.modname)
-    return module.resolve(self.real_name(name))
-
-From.self_resolve = from_self_resolve
-
-def global_self_resolve(node, name, _):
-    """self resolve on Global is equivalent to resolve at the module level"""
-    return node.root().resolve(name)
-Global.self_resolve = global_self_resolve
-
-def name_self_resolve(node, _, frame):
-    """self resolve on Name is equivalent to resolve the node name instead
-    of the one given as argument
+def if_block_range(node, lineno):
+    """handle block line numbers range for if/elif statements
     """
-    return frame.resolve(node.name)
-Name.self_resolve = name_self_resolve
+    last = None
+    for test, testbody in node.tests[1:]:
+        if lineno == testbody.source_line():
+            return lineno, lineno
+        if lineno <= testbody.last_source_line():
+            return lineno, testbody.last_source_line()
+        if last is None:
+            last = testbody.source_line() - 1
+    return elsed_block_range(node, lineno, last)
 
-def getattr_self_resolve(node, _, frame):
-    """self resolve on Getattr is equivalent to resolve at the parent frame"""
-    child = node.getChildNodes()[0]
-    if not isinstance(child, (Name, Getattr)):
-        # FIXME: dunno what to do here
-        raise ResolveError(_)
-    return child.self_resolve(None, frame).resolve(node.attrname)
-Getattr.self_resolve = getattr_self_resolve
+If.block_range = if_block_range
 
+def try_except_block_range(node, lineno):
+    """handle block line numbers range for try/except statements
+    """
+    last = None
+    for excls, exinst, exbody in node.handlers:
+        if excls and lineno == excls.source_line():
+            return lineno, lineno
+        if exbody.source_line() <= lineno <= exbody.last_source_line():
+            return lineno, exbody.last_source_line()
+        if last is None:
+            last = exbody.source_line() - 1
+    return elsed_block_range(node, lineno, last)
+
+TryExcept.block_range = try_except_block_range
+
+def elsed_block_range(node, lineno, last=None):
+    """handle block line numbers range for try/finally, for and while
+    statements
+    """
+    if lineno == node.source_line():
+        return lineno, lineno
+    if node.else_:
+        if lineno >= node.else_.source_line():
+            return lineno, node.else_.last_source_line()
+        return lineno, node.else_.source_line() - 1
+    return lineno, last or node.last_source_line()
+
+TryFinally.block_range = elsed_block_range
+While.block_range = elsed_block_range
+For.block_range = elsed_block_range
 
 # From and Import #############################################################
 
@@ -360,20 +310,6 @@ From.real_name = real_name
 Import.real_name = real_name
 
 
-# AssName #####################################################################
-
-def rhs(node):
-    """return the node corresponding to the assigned value (right hand side)
-    """
-    # FIXME AssTuple...
-    assnode = node.parent
-    while assnode and not isinstance(assnode, Assign):
-#        assert assnode is not node.parent, assnode
-        assnode = assnode.parent
-    return assnode and assnode.expr or None
-AssName.rhs = rhs
-AssAttr.rhs = rhs
-        
 # as_string ###################################################################
 
 def add_as_string(node):
@@ -478,8 +414,9 @@ Class.as_string = class_as_string
 
 def compare_as_string(node):
     """return an ast.Compare node as string"""
-    rhs = ' '.join(['%s %s' % (op, expr.as_string()) for op, expr in node.ops])
-    return '%s %s' % (node.expr.as_string(), rhs)
+    rhs_str = ' '.join(['%s %s' % (op, expr.as_string())
+                        for op, expr in node.ops])
+    return '%s %s' % (node.expr.as_string(), rhs_str)
 Compare.as_string = compare_as_string
 
 def const_as_string(node):
@@ -518,6 +455,10 @@ def ellipsis_as_string(node):
     return '...'
 Ellipsis.as_string = ellipsis_as_string
 
+def empty_as_string(node):
+    return ''
+EmptyNode.as_string = empty_as_string
+
 def exec_as_string(node):
     """return an ast.Exec node as string"""
     if node.globals:
@@ -525,7 +466,8 @@ def exec_as_string(node):
                                       node.locals.as_string(),
                                       node.globals.as_string())
     if node.locals:
-        return 'exec %s in %s' % (node.expr.as_string(), node.locals.as_string())
+        return 'exec %s in %s' % (node.expr.as_string(),
+                                  node.locals.as_string())
     return 'exec %s' % node.expr.as_string()
 Exec.as_string = exec_as_string
 
@@ -552,6 +494,29 @@ def function_as_string(node):
                                       node.code.as_string())
 Function.as_string = function_as_string
 
+def genexpr_as_string(node):
+    """return an ast.GenExpr node as string"""
+    return '(%s)' % node.code.as_string()
+GenExpr.as_string = genexpr_as_string
+
+def genexprinner_as_string(node):
+    """return an ast.GenExpr node as string"""
+    return '%s %s' % (node.expr.as_string(), ' '.join([n.as_string()
+                                                       for n in node.quals]))
+GenExprInner.as_string = genexprinner_as_string
+
+def genexprfor_as_string(node):
+    """return an ast.GenExprFor node as string"""
+    return 'for %s in %s %s' % (node.assign.as_string(),
+                                node.iter.as_string(),
+                                ' '.join([n.as_string() for n in node.ifs]))
+GenExprFor.as_string = genexprfor_as_string
+
+def genexprif_as_string(node):
+    """return an ast.GenExprIf node as string"""
+    return 'if %s' % node.test.as_string()
+GenExprIf.as_string = genexprif_as_string
+
 def getattr_as_string(node):
     """return an ast.Getattr node as string"""
     return '%s.%s' % (node.expr.as_string(), node.attrname)
@@ -786,11 +751,6 @@ def yield_as_string(node):
     return 'yield %s' % node.value.as_string()
 Yield.as_string = yield_as_string
 
-# to backport into compiler ###################################################
-
-EmptyNode.getChildNodes = lambda self: ()
-
-# private utilities ###########################################################
 
 def _import_string(names):
     """return a list of (name, asname) formatted as a string
@@ -802,3 +762,7 @@ def _import_string(names):
         else:
             _names.append(name)
     return  ', '.join(_names)
+
+# to backport into compiler ###################################################
+
+EmptyNode.getChildNodes = lambda self: ()
