@@ -18,31 +18,33 @@ Each new methods and attributes added on each class are documented
 below.
 
 
-:version:   $Revision: 1.6 $  
+:version:   $Revision: 1.28 $  
 :author:    Sylvain Thenault
-:copyright: 2003-2005 LOGILAB S.A. (Paris, FRANCE)
+:copyright: 2003-2006 LOGILAB S.A. (Paris, FRANCE)
 :contact:   http://www.logilab.fr/ -- mailto:python-projects@logilab.org
-:copyright: 2003-2005 Sylvain Thenault
+:copyright: 2003-2006 Sylvain Thenault
 :contact:   mailto:thenault@gmail.com
 """
 
 from __future__ import generators
 
-__revision__ = "$Id: scoped_nodes.py,v 1.6 2005/11/02 11:56:45 syt Exp $"
+__revision__ = "$Id: scoped_nodes.py,v 1.28 2006-04-20 07:37:28 syt Exp $"
 __doctype__ = "restructuredtext en"
 
 import sys
 from compiler.ast import Module, Class, Function, Lambda, Dict, Tuple, Raise, \
-     Pass, Raise, Return
+     Pass, Raise, Yield, Name, Const
 
 from logilab.common.compat import chain        
 
-from logilab.astng._exceptions import ResolveError, NotFoundError, NoDefault
+from logilab.astng._exceptions import NotFoundError, NoDefault, \
+     ASTNGBuildingException, InferenceError
 from logilab.astng.utils import extend_class
+from logilab.astng import YES, MANAGER, Instance, unpack_infer, _infer_stmts
 
 # module class dict/iterator interface ########################################
-
-class LocalsDictMixIn:
+    
+class LocalsDictMixIn(object):
     """ this class provides locals handling common to Module, Function
     and Class nodes, including a dict like interface for direct access
     to locals information
@@ -59,6 +61,7 @@ class LocalsDictMixIn:
     # value    
     locals = None
     
+    
     def frame(self):
         """return the first node defining a new local scope (i.e. Module,
         Function or Class)
@@ -67,7 +70,8 @@ class LocalsDictMixIn:
     
     def set_local(self, name, stmt):
         """define <name> in locals (<stmt> is the node defining the name)
-        if the node is a Module node (i.e. has globals), add the name to globals
+        if the node is a Module node (i.e. has globals), add the name to
+        globals
 
         if the name is already defined, ignore it
         """
@@ -96,13 +100,14 @@ class LocalsDictMixIn:
         return self.locals[item][0]
     
     def __iter__(self):
-        """method from the `dict` interface returning an iterator on `self.keys()`
+        """method from the `dict` interface returning an iterator on
+        `self.keys()`
         """
         return iter(self.keys())
     
     def keys(self):
-        """method from the `dict` interface returning a tuple containing locally
-        defined names
+        """method from the `dict` interface returning a tuple containing
+        locally defined names
         """
         return self.locals.keys()
 ##         associated to nodes which are instance of `Function` or
@@ -113,27 +118,28 @@ class LocalsDictMixIn:
 ##             return self.__keys
 ##         except AttributeError:
 ##             keys = [member.name for member in self.locals.values()
-##                     if (isinstance(member, Function) or isinstance(member, Class))
+##                     if (isinstance(member, Function)
+##                         or isinstance(member, Class))
 ##                         and member.parent.frame() is self]
 ##             self.__keys = tuple(keys)
 ##             return keys
 
     def values(self):
-        """method from the `dict` interface returning a tuple containing locally
-        defined nodes which are instance of `Function` or `Class`
+        """method from the `dict` interface returning a tuple containing
+        locally defined nodes which are instance of `Function` or `Class`
         """
         return [self[key] for key in self.keys()]
     
     def items(self):
-        """method from the `dict` interface returning a list of tuple containing
-        each locally defined name with its associated node, which is an instance of
-        `Function` or `Class`
+        """method from the `dict` interface returning a list of tuple
+        containing each locally defined name with its associated node,
+        which is an instance of `Function` or `Class`
         """
         return zip(self.keys(), self.values())
 
     def has_key(self, name):
-        """method from the `dict` interface returning True if the given name
-        is defined in the locals dictionary
+        """method from the `dict` interface returning True if the given
+        name is defined in the locals dictionary
         """
         return self.locals.has_key(name)
     
@@ -144,10 +150,25 @@ extend_class(Class, LocalsDictMixIn)
 extend_class(Function, LocalsDictMixIn)
 extend_class(Lambda, LocalsDictMixIn)
 
+class GetattrMixIn(object):
+    def getattr(self, name, path=None):
+        try:
+            return self.locals[name]
+        except KeyError:
+            raise NotFoundError(name)
+        
+    def igetattr(self, name, path=None):
+        """infered getattr"""
+        try:
+            return _infer_stmts(self.getattr(name, path), name, frame=self, path=path)
+        except NotFoundError:
+            raise InferenceError(name)
+extend_class(Module, GetattrMixIn)
+extend_class(Class, GetattrMixIn)
 
 # Module  #####################################################################
 
-class ModuleNG:
+class ModuleNG(object):
     """/!\ this class should not be used directly /!\ it's
     only used as a methods and attribute container, and update the
     original class from the compiler.ast module using its dictionnary
@@ -169,6 +190,19 @@ class ModuleNG:
     # as value
     globals = None
   
+    def getattr(self, name, path=None):
+        try:
+            return self.locals[name]
+        except KeyError:
+            if self.package:
+                try:
+                    return [self.import_module(name, relative_only=True)]
+                except KeyboardInterrupt:
+                    raise
+                except:
+                    pass
+            raise NotFoundError(name)
+        
     def _append_node(self, child_node):
         """append a child version specific to Module node"""
         self.node.nodes.append(child_node)
@@ -178,12 +212,36 @@ class ModuleNG:
         """return the source line number, 0 on a module"""
         return 0
 
+    def fully_defined(self):
+        """return True if this module has been built from a .py file
+        and so contains a complete representation including the code
+        """
+        return self.file is not None and self.file.endswith('.py')
+    
     def statement(self):
         """return the first parent node marked as statement node
         consider a module as a statement...
         """
         return self
 
+    def import_module(self, modname, relative_only=False):
+        """import the given module considering self as context"""
+        try:
+            return MANAGER.astng_from_module_name(self.relative_name(modname))
+        except ASTNGBuildingException:
+            if relative_only:
+                raise
+        module = MANAGER.astng_from_module_name(modname)
+        return module
+
+    def relative_name(self, modname):
+        if self.package:
+            return '%s.%s' % (self.name, modname)
+        package_name = '.'.join(self.name.split('.')[:-1])
+        if package_name:
+            return '%s.%s' % (package_name, modname)
+        return modname
+        
     def wildcard_import_names(self):
         """return the list of imported names when this module is 'wildard
         imported'
@@ -202,11 +260,15 @@ class ModuleNG:
             except AttributeError:
                 return [name for name in living.__dict__.keys()
                         if not name.startswith('_')]
+        # else lookup the astng
         try:
-            explicit = self['__all__'].rhs()
+            explicit = self['__all__'].assigned_stmts().next()
             # should be a tuple of constant string
             return [const.value for const in explicit.nodes]
-        except KeyError:
+        except (KeyError, AttributeError, InferenceError):
+            # XXX should admit we have lost if there is something like
+            # __all__ that we've not been able to analyse (such as
+            # dynamically constructed __all__)
             return [name for name in self.keys()
                     if not name.startswith('_')]
 
@@ -214,7 +276,7 @@ extend_class(Module, ModuleNG)
 
 # Function  ###################################################################
 
-class FunctionNG:
+class FunctionNG(object):
     """/!\ this class should not be used directly /!\ it's
     only used as a methods and attribute container, and update the
     original class from the compiler.ast module using its dictionnary
@@ -240,31 +302,26 @@ class FunctionNG:
         """
         for child_node in self.code.getChildNodes():
             if isinstance(child_node, Raise) and child_node.expr1:
-                names = child_node.expr1.names()
-                if names and names[0] == 'NotImplementedError':
-                    return True
+                try:
+                    name = child_node.expr1.nodes_of_class(Name).next()
+                    if name.name == 'NotImplementedError':
+                        return True
+                except StopIteration:
+                    pass
             if pass_is_abstract and isinstance(child_node, Pass):
                 return True
             return False
         # empty function is the same as function with a single "pass" statement
         if pass_is_abstract:
             return True
-        
-    def self_resolve(self, name, frame):
-        """overriden from NodeNG class:
-        self resolve on Function check we did not catch a function's argument
-        """
-        if self is frame:
-            # the searched name has been found in the function locals,
-            # and resolved to the function itself, so it must be one
-            # of the function's arguments
-            try:
-                value = self.default_value(name)
-            except NoDefault:
-                raise ResolveError(name)
-            return value.self_resolve(name, self.parent)
-        return self
 
+    def is_generator(self):
+        """return true if this is a generator function"""
+        try:
+            return self.nodes_of_class(Yield, skip_klass=Function).next()
+        except StopIteration:
+            return False
+        
     def format_args(self):
         """return arguments formatted as string"""
         if self.argnames is None: # information is missing
@@ -280,7 +337,8 @@ class FunctionNG:
             elif args and i == last or (kwargs and i == last - 1):
                 name = '*%s' % name
             elif i >= default_idx:
-                name = '%s=%s' % (name, self.defaults[i - default_idx].as_string())
+                default_str = self.defaults[i - default_idx].as_string()
+                name = '%s=%s' % (name, default_str)
             result.append(name)
         return ', '.join(result)
 
@@ -291,25 +349,28 @@ class FunctionNG:
         """
         if self.argnames is None: # information is missing
             raise NoDefault()
-        args, kwargs, last, default_idx = self._pos_information()
-        for i in range(len(self.argnames)):
-            name = self.argnames[i]
-            if name != argname:
-                continue
-            if i == last and kwargs:
-                val_node = Dict([])
-                val_node.parent = self
-                return val_node
-            if name == argname and ((args and i == last)
-                                    or (kwargs and i == last - 1)):
-                val_node = Tuple([])
-                val_node.parent = self
-                return val_node
-            if i >= default_idx:
-                return self.defaults[i - default_idx]
-            break
+        args, kwargs, last, defaultidx = self._pos_information()
+        i = self.argnames.index(argname)
+        if i >= defaultidx and (i - defaultidx) < len(self.defaults):
+            return self.defaults[i - defaultidx]
         raise NoDefault()
 
+    def mularg_class(self, argname):
+        """if the given argument is a * or ** argument, return respectivly
+        a Tuple or Dict instance, else return None
+        """
+        args, kwargs, last, defaultidx = self._pos_information()
+        i = self.argnames.index(argname)
+        if i == last and kwargs:
+            valnode = Dict([])
+            valnode.parent = self
+            return valnode
+        if args and (i == last or (kwargs and i == last - 1)):
+            valnode = Tuple([])
+            valnode.parent = self
+            return valnode
+        return None
+
     def _pos_information(self):
         """return a 4-uple with positional information about arguments:
         (true if * is used,
@@ -320,25 +381,10 @@ class FunctionNG:
         args = self.flags & 4
         kwargs = self.flags & 8
         last = len(self.argnames) - 1
-        default_idx = len(self.argnames) - (len(self.defaults) +
-                                            (args and 1 or 0) +
-                                            (kwargs and 1 or 0))
-        return args, kwargs, last, default_idx
-
-    # FIXME: are methods below really useful ?
-    
-##     def raises(self):
-##         """return an iterator on exceptions raised below the given node
-##         (ie expr1 attribute of the Raise node)
-##         """
-##         for child in self.nodes_of_class(Raise):
-##             if child.expr1:
-##                 yield child.expr1
-##     def returns(self):
-##         """return an iterator on nodes used in return statements below the given
-##         node"""
-##         for child in self.nodes_of_class(Return):
-##             yield child.value
+        defaultidx = len(self.argnames) - (len(self.defaults) +
+                                           (args and 1 or 0) +
+                                           (kwargs and 1 or 0))
+        return args, kwargs, last, defaultidx
 
 extend_class(Function, FunctionNG)
 
@@ -346,6 +392,7 @@ extend_class(Function, FunctionNG)
 Lambda._pos_information = FunctionNG._pos_information.im_func
 Lambda.format_args = FunctionNG.format_args.im_func
 Lambda.default_value = FunctionNG.default_value.im_func
+Lambda.mularg_class = FunctionNG.mularg_class.im_func
 Lambda.type = 'function'
 
 # Class ######################################################################
@@ -371,16 +418,26 @@ def _class_type(klass):
         klass._type = 'class'
     return klass._type
 
+def _class_newstyle(klass):
+    """return a if the given class is new-style or not
+    """
+    if klass._newstyle is not None:
+        return klass._newstyle
+    for base in klass.ancestors(recurs=False):
+        if base.newstyle:
+            klass._newstyle = base.newstyle
+            break
+    if klass._newstyle is None:
+        klass._newstyle = False
+    return klass._newstyle
+
 def _iface_hdlr(iface_node):
     """a handler function used by interfaces to handle suspicious
     interface nodes
     """
-    try:
-        yield iface_node.resolve_dotted(iface_node.as_string())
-    except:
-        return
+    return True
 
-class ClassNG:
+class ClassNG(object):
     """/!\ this class should not be used directly /!\ it's
     only used as a methods and attribute container, and update the
     original class from the compiler.ast module using its dictionnary
@@ -392,6 +449,11 @@ class ClassNG:
                     doc="class'type, possible values are 'class' | "
                     "'metaclass' | 'interface' | 'exception'")
     
+    _newstyle = None
+    newstyle = property(_class_newstyle,
+                        doc="boolean indicating if it's a new style class"
+                        "or not")
+    
     # attributes below are set by the builder module or by raw factories
     
     # a dictionary of class instances attributes
@@ -400,7 +462,7 @@ class ClassNG:
     # in the class definition)
     basenames = None
 
-    def ancestors(self, recurs=True):
+    def ancestors(self, recurs=True, path=None):
         """return an iterator on the node base classes in a prefixed
         depth first order
         
@@ -409,32 +471,41 @@ class ClassNG:
           ancestors only
         """
         # FIXME: should be possible to choose the resolution order
-        for baseobj in self.parent.frame().resolve_all(self.basenames):
-            if not isinstance(baseobj, Class):
-                # duh ?
+        if path is None:
+            path = []
+        for stmt in self.bases:
+            try:
+                for baseobj in stmt.infer(path=path):
+                    if not isinstance(baseobj, Class):
+                        # duh ?
+                        continue
+                    yield baseobj
+                    if recurs:
+                        for grandpa in baseobj.ancestors(True, path):
+                            yield grandpa
+            except InferenceError:
+                #import traceback
+                #traceback.print_exc()
+                # XXX log error ?
                 continue
-            yield baseobj
-            if recurs:
-                for grandpa in baseobj.ancestors(True):
-                    yield grandpa
-
-    def local_attr_ancestors(self, name):
+            
+    def local_attr_ancestors(self, name, path=None):
         """return an iterator on astng representation of parent classes
         which have <name> defined in their locals
         """
-        for astng in self.ancestors():
+        for astng in self.ancestors(path=path):
             if astng.locals.has_key(name):
                 yield astng
 
-    def instance_attr_ancestors(self, name):
+    def instance_attr_ancestors(self, name, path=None):
         """return an iterator on astng representation of parent classes
         which have <name> defined in their instance attribute dictionary
         """
-        for astng in self.ancestors():
+        for astng in self.ancestors(path=path):
             if astng.instance_attrs.has_key(name):
                 yield astng
 
-    def local_attr(self, name):
+    def local_attr(self, name, path=None):
         """return the astng associated to name in this class locals or
         in its parents
 
@@ -446,12 +517,12 @@ class ClassNG:
             return self[name]
         except KeyError:
             # get if from the first parent implementing it if any
-            for class_node in self.local_attr_ancestors(name):
+            for class_node in self.local_attr_ancestors(name, path):
                 return class_node[name]
         raise NotFoundError(name)
         
-    def instance_attr(self, name):
-        """return the astng associated to name in this class instance
+    def instance_attr(self, name, path=None):
+        """return the astng nodes associated to name in this class instance
         attributes dictionary or in its parents
 
         :raises `NotFoundError`:
@@ -462,16 +533,67 @@ class ClassNG:
             return self.instance_attrs[name]
         except KeyError:
             # get if from the first parent implementing it if any
-            for class_node in self.instance_attr_ancestors(name):
+            for class_node in self.instance_attr_ancestors(name, path):
                 return class_node.instance_attrs[name]
         raise NotFoundError(name)
 
-    def attr(self, name):
+    def getattr(self, name, path=None):
+        """this method doesn't look in the instance_attrs dictionary since it's
+        done by an Instance proxy at inference time.
+        
+        It may return a YES object if the attribute has not been actually
+        found but a __getattr__ or __getattribute__ method is defined
+        """
+        if name in self.locals:
+            return self.locals[name]
+        for classnode in self.ancestors(recurs=False, path=path):
+            try:
+                return classnode.getattr(name, path)
+            except NotFoundError:
+                continue
+        raise NotFoundError(name)
+
+    def igetattr(self, name, path=None):
+        """infered getattr, need special treatment in class to handle
+        descriptors
+        """
         try:
-            return self.instance_attr(name)
+            for infered in _infer_stmts(self.getattr(name, path), name,
+                                        frame=self, path=path):
+                # yield YES object instead of descriptors when necessary
+                if not isinstance(infered, Const) and isinstance(infered, Instance):
+                    try:
+                        infered._proxied.getattr('__get__', path)
+                    except NotFoundError:
+                        yield infered
+                    else:
+                        yield YES
+                else:
+                    yield infered
         except NotFoundError:
-            return self.local_attr(name)
-            
+            if not name.startswith('__') and self.has_dynamic_getattr(path):
+                # class handle some dynamic attributes, return a YES object
+                yield YES
+            else:
+                raise InferenceError(name)
+        
+    def has_dynamic_getattr(self, path=None):
+        """return True if the class has a custom __getattr__ or
+        __getattribute__ method
+        """
+        try:
+            self.getattr('__getattr__', path)
+            return True
+        except NotFoundError:
+            try:
+                getattribute = self.getattr('__getattribute__', path)[0]
+                if getattribute.root().name != '__builtin__':
+                    # class has a custom __getattribute__ defined
+                    return True
+            except NotFoundError:
+                pass
+        return False
+    
     def methods(self):
         """return an iterator on all methods defined in the class and
         its ancestors
@@ -489,32 +611,34 @@ class ClassNG:
         for member in self.values():
             if isinstance(member, Function):
                 yield member
-
+                
     def interfaces(self, herited=True, handler_func=_iface_hdlr):
         """return an iterator on interfaces implemented by the given
         class node
         """
         # FIXME: what if __implements__ = (MyIFace, MyParent.__implements__)...
         try:
-            implements = self.local_attr('__implements__')
+            implements = Instance(self).getattr('__implements__')[0]
         except NotFoundError:
             return
         if not herited and not implements.frame() is self:
             return
-        implements = implements.rhs()
-        if hasattr(implements, 'nodes'):
-            implements = implements.nodes
-        else:
-            implements = (implements,)
-    ##     if not (isinstance(implements, tuple) or isinstance(implements, list)):
-    ##         implements = (implements,)
-        for iface in implements:
-    ##         if isinstance(iface, Class):
-    ##             yield iface
-    ##             continue
-            # let the handler function take care of this....
-            for iface in handler_func(iface):
+        oneinf = False
+        for iface in unpack_infer(implements):
+            if iface is YES:
+                continue
+            if handler_func(iface):
+                oneinf = True
                 yield iface
-
+        if not oneinf:
+            raise InferenceError()
+##         if hasattr(implements, 'nodes'):
+##             implements = implements.nodes
+##         else:
+##             implements = (implements,)
+##         for iface in implements:
+##             # let the handler function take care of this....
+##             for iface in handler_func(iface):
+##                 yield iface
 
 extend_class(Class, ClassNG)
